/*
 * Copyright (c) 2005-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 177 $
 * $LastChangedDate: 2007-05-07 11:38:46 +0200 (Mon, 07 May 2007) $
 * 
 */

package com.codedroids.forms;

import java.util.*;
import org.w3c.dom.*;


/**
 * Children:<br>
 *   ident<br>
 *   name<br>
 *   label<br>
 *   validators<br>
 *   layout<br>
 *   readonly<br>
 *   ps<br>
 *   format<br> 
 *   render<br> 
 * Formatter object that takes locale, number/date format strings and returns date, numbers formatted as strings
 * This class does not hold the value as some needs one and other needs multiple values, see subclasses.
 */
public abstract class A_Field implements I_Field {

	protected Form form;
	private List errorMessages;

	private String tagName;
	private String ident;
	private String name;
	private HashMap labels;
	private Boolean readOnly;
	private String render;
	private List validators;
	private Layout layout;
	private HashMap postscripts;
	private FieldFormat format; // currently not used by render

	public A_Field(Form form) {
		super();
		this.form = form;
	}
	
	public A_Field(Form form, Node field) {
		super();
		this.form = form;
		this.populateFromNode(field);
	}
	
	abstract public int cardinality();
	abstract public List getValues();
	abstract public void setValues(List values);
	abstract public void setValues(String[] values);
	abstract public List getOptions();
	abstract public void setOptions(List opts);
	
    public String toHTML() throws UnknownFieldRenderException {
		return form().getFieldRender(getRender()).render(this);
    }

	public int validate() {
		if(getValidators() != null) {
			Iterator it = getValidators().iterator();
			while(it.hasNext()) {
				I_Validator validator = (I_Validator)it.next();
				try {
					if(!validator.validate(this)) {
						String msg = getLabel();
						if(msg == null)
							msg = getName();
						addErrorMessage(msg + ": " + validator.getErrorMessage());
						return 1; // unlikely that it makes sense to check further validators
					}
				} catch(Exception e) {
					String msg = getLabel();
					if(msg == null)
						msg = getName();
					addErrorMessage(msg + ": " + e.getMessage());
					return 1;
				}
					
			}
		}
		return 0;
	}
    
	public String tagName() {
		return tagName;
	}
	
	public Form form() {
		return form;
	}
	
	public boolean isReadOnly() {
		if(readOnly != null)
			return readOnly.booleanValue();
		else
			return form.isReadOnly();
	}
	
	public List getErrorMessages() {
		return errorMessages;
	}
	
	public void clearErrorMessages() {
		errorMessages = null;
	}
	public void addErrorMessage(String msg) {
		if(errorMessages == null)
			errorMessages = new ArrayList();
		errorMessages.add(msg);
	}
	
	/**
	 * This handles attributes and typical children of a node. Subclass if this is not sufficient.
     * Generally it should be sufficient to subclass handleAttribute and handleChildNode methods.
	 * @param field
	 */
	public void populateFromNode(Node field) 
	{
		this.tagName = field.getNodeName();
		/** No attributes for now 
		NamedNodeMap attributes = field.getAttributes();
		int len = attributes.getLength();
		for(int i = 0; i < len; i++) {
			Node attr = attributes.item(i);
			short at = attr.getNodeType();
			String an = attr.getNodeName();
			handleAttribute(attr, at, an);
		}
		**/
		NodeList children = field.getChildNodes();
		int len = children.getLength();
		for(int i = 0; i < len; i++) {
			Node child = children.item(i);
			short ct = child.getNodeType();
			String cn = child.getNodeName();
			handleChildNode(child, ct, cn);
		}
	}
	
	/**
	 * The optional ident and style are common between all fields
	 * @param attr
	 * @param attrType
	 * @param attrName
	 * @return true if attr is considered handled, false if it didn't cause any changes
	 */
	/* No attributes...
	public boolean handleAttribute(Node attr, short attrType, String attrName)
	{
		if(attrType == Node.ATTRIBUTE_NODE) {
			if(attrName.equals(OCForm.FBATTR_IDENT)) {
				setIdent(attr.getNodeValue());
				return true;
			}
			else if(attrName.equals(OCForm.FBATTR_STYLE)) {
				setStyle(attr.getNodeValue());
				return true;
			}
		}
		return false;
	}
	**/
	
	/**
	 * Pick up ident and style if present. Subclasses must should call super
	 * @param child
	 * @param childType
	 * @param childName
	 * @return true if child is considered handled, false if it didn't cause any changes
	 */
	public boolean handleChildNode(Node child, short childType, String childName)
	{
		if(childType == Node.ELEMENT_NODE) {
			//com.opencms.core.A_OpenCms.log(com.opencms.boot.I_CmsLogChannels.C_OPENCMS_INFO, "[A_Field] handle child = " + childName);
			if(childName.equals(Form.FBTAG_IDENT)) {
				setIdent(Form.collectText(child));
				return true;
			}
			else if(childName.equals(Form.FBTAG_NAME)) {
				setName(Form.collectText(child));
				return true;
			}
			else if(childName.equals(Form.FBTAG_LABEL)) {
				String v = Form.collectText(child);
				String l = ""; // in case it is not specified 
				NamedNodeMap attrs = child.getAttributes();
				for(int i=0; i<attrs.getLength(); i++) {
					Node attr = attrs.item(i);
					String an = attr.getNodeName();
					if(an.equals(Form.FBATTR_LOCALE)) {
						l = Form.collectText(attr);
					}
				}
				setLabel(v, l);
				return true;
			}
			else if(childName.equals(Form.FBTAG_READONLY)) {
				setReadOnly(Form.collectText(child));
				return true;
			}
			else if(childName.equals(Form.FBTAG_RENDER)) {
				setRender(Form.collectText(child));
				return true;
			}
			else if (childName.equals(Form.FBTAG_VALIDATORS)) {
				parseValidators(child);
				return true;
			}
			else if (childName.equals(Form.FBTAG_LAYOUT)) {
				setLayout(new Layout(form, child));
				return true;
			}
			else if(childName.equals(Form.FBTAG_POSTSCRIPT)) {
				String v = Form.collectText(child);
				String l = ""; // in case it is not specified 
				NamedNodeMap attrs = child.getAttributes();
				for(int i=0; i<attrs.getLength(); i++) {
					Node attr = attrs.item(i);
					String an = attr.getNodeName();
					if(an.equals(Form.FBATTR_LOCALE)) {
						l = Form.collectText(attr);
					}
				}
				setPostscript(v, l);
				return true;
			}
			else if (childName.equals(Form.FBTAG_FORMAT)) {
				setFormat(new FieldFormat(form, child));
				return true;
			}
		}
		return false;
	}

	
	private void parseValidators(Node validators)
	{
		NodeList nl = validators.getChildNodes();
		int len = nl.getLength();
		for(int i = 0; i < len; i++) {
			Node validator = nl.item(i);
			short ft = validator.getNodeType();
			switch(ft) {
				case Node.ELEMENT_NODE:
					String fn = validator.getNodeName();
					if (fn.equals(Form.FBTAG_REQUIRED)) {
						addValidator(new RequiredValidator(form, validator));
					}
					else if (fn.equals(Form.FBTAG_EMAIL)) {
						addValidator(new EmailValidator(form, validator));
					}
					else if (fn.equals(Form.FBTAG_REGEX)) {
						addValidator(new RegExValidator(form, validator));
					}
					else if (fn.equals(Form.FBTAG_OPTIONS)) {
						addValidator(new OptionsValidator(form, validator));
					}
					break;
				default:		
					break;
			}
		}
	}
	
	//
	// Accessors
	//
		
	public String getIdent() {
		if(ident == null)
			return name; // should not harm in general and certain dynamic components requires an id for the field
		else
			return ident;
	}
	
	public void setIdent(String str) {
		this.ident = str;
	}
	
	public String getName() {
		if(name == null)
			return getIdent(); // since name is kind of deprecated in general we'll use this as default so one need not provide a name
		else
			return name;
	}
	
	public void setName(String str) {
		this.name = str;
	}
	
	/**
	 * First it checks the label based on the current locale of the form, if that fails
	 * it tries one more time with the "" locale (i.e. labels not specifically marked with locale
	 * @return
	 */
	public String getLabel() {
		String retval = getLabel(form.getLocale());
		if(retval==null && labels!=null)
			retval = (String)labels.get("");
		return retval;
	}
	
	/**
	 * Sets the label with un-specified locale
	 * @param str
	 */
	public void setLabel(String str) {
		setLabel(str, "");
	}

	/**
	 * Allows explicit get according to locale, primarily for reading/writing to XML representation
	 * @param locale
	 * @return
	 */
	public String getLabel(String locale) {
		if(labels != null) {
			return (String)labels.get(locale);
		} else {
			return null;
		}
	}
	/**
	 * Allows explicit set according to locale, primarily for reading/writing to XML representation
	 * @param str
	 * @param locale
	 */
	public void setLabel(String str, String locale) {
		if(labels == null)
			labels = new HashMap();
		labels.put(locale, str);
	}

	public String getReadOnly() {
		return ((readOnly != null && readOnly.booleanValue()) ? "true" : "false");
	}
	
	public void setReadOnly(String str) {
		if(str == null || str.length()==0) {
			this.readOnly = Boolean.FALSE; // empty field means activate it
		} else {
			try {
				this.readOnly = Boolean.valueOf(str);
			} catch (Exception e) {
				this.readOnly = Boolean.FALSE;
			}
		}
	}
	
	public void setReadOnly(Boolean b) {
		this.readOnly = b;
	}

	public void setReadOnly(boolean b) {
		this.readOnly = Boolean.valueOf(b);
	}

	public String getRender() {
		if(render == null)
			return "default";
		else
			return render;
	}
	
	public void setRender(String str) {
		this.render = str;
	}

	public Layout getLayout() {
		if(layout == null && form != null)
			return form.getLayout();
		else
			return layout;
	}
	
	public void setLayout(Layout h) {
		this.layout = h;
	}
	
	/**
	 * First it checks the postscript based on the current locale of the form, if that fails
	 * it tries one more time with the "" locale (i.e. labels not specifically marked with locale
	 * @return
	 */
	public String getPostscript() {
		String retval = getPostscript(form.getLocale());
		if(retval==null && postscripts!=null)
			retval = (String)postscripts.get("");
		return retval;
	}
	
	public void setPostscript(String str) {
		setPostscript(str, "");
	}

	public String getPostscript(String locale) {
		if(postscripts != null) {
			return (String)postscripts.get(locale);
		} else {
			return null;
		}
	}

	public void setPostscript(String str, String locale) {
		if(postscripts == null)
			postscripts = new HashMap();
		postscripts.put(locale, str);
	}
	
	public FieldFormat getFormat() {
		return format;
	}
	
	public void setFormat(FieldFormat format) {
		this.format = format;
	}
	
	public List getValidators() {
		return validators;
	}

	public void setSetValidators(List str) {
		validators = str;
	}
	
	public void addValidator(I_Validator f) {
		if(validators == null)
			validators = new ArrayList();
		validators.add(f);
	}
	
	public void removeValidator(I_Validator f) {
		if(validators != null)	
			validators.remove(f);
	}
	
	//
	//
	//
	
	public String toString() {
		return toXML("", true);
	}
	
	/**
	 * Shouldn't be any reason to subclass this
	 * @param indent
	 * @return
	 */
	public String toXML(String indent, boolean inclVal) {
		String tmp = childrenToXML(indent + Form.PP_INDENT, inclVal);
		if(tmp != null)
			return "\n" + indent + "<"+tagName()+">" + tmp + "\n" + indent + "</"+tagName()+">";
		else
			return "\n" + indent + "<"+tagName()+" />";
	}
	
	/**
	 * Subclasses should start by calling the super to get it's instance vars in the output
	 * @param indent
	 * @return
	 */
	public String childrenToXML(String indent, boolean inclVal) {
		StringBuffer b = new StringBuffer();
		if(getIdent() != null)
			b.append("\n" + indent + "<"+Form.FBTAG_IDENT+">" + getIdent() + "</"+Form.FBTAG_IDENT+">");
 		if(getName() != null)
			b.append("\n" + indent + "<"+Form.FBTAG_NAME+">" + getName() + "</"+Form.FBTAG_NAME+">");
		if(labels != null) {
			Iterator it = labels.keySet().iterator();
			while(it.hasNext()) {
				String loc = (String)it.next();
				if("".equals(loc)) {
					b.append("\n" + indent + "<"+Form.FBTAG_LABEL+">" + labels.get(loc) + "</"+Form.FBTAG_LABEL+">");
				} else {
					b.append("\n" + indent + "<"+Form.FBTAG_LABEL+" "+Form.FBATTR_LOCALE+"=\"" +loc+ "\">" + labels.get(loc) + "</"+Form.FBTAG_LABEL+">");
				}
			}
		}
		if(isReadOnly())
			b.append("\n" + indent + "<"+Form.FBTAG_READONLY+" />");
		if(postscripts != null) {
			Iterator it = postscripts.keySet().iterator();
			while(it.hasNext()) {
				String loc = (String)it.next();
				if("".equals(loc)) {
					b.append("\n" + indent + "<"+Form.FBTAG_POSTSCRIPT+">" + labels.get(loc) + "</"+Form.FBTAG_POSTSCRIPT+">");
				} else {
					b.append("\n" + indent + "<"+Form.FBTAG_POSTSCRIPT+" "+Form.FBATTR_LOCALE+"=\"" +loc+ "\">" + labels.get(loc) + "</"+Form.FBTAG_POSTSCRIPT+">");
				}
			}
		}
//		if(getPostscript() != null)
//			b.append("\n" + indent + "<"+Form.FBTAG_POSTSCRIPT+">" + getPostscript() + "</"+Form.FBTAG_POSTSCRIPT+">");
		if(inclVal && getValidators() != null) {
			b.append("\n" + indent + "<"+Form.FBTAG_VALIDATORS+">");
			Iterator it = getValidators().iterator();
			while(it.hasNext()){
				I_Validator v = (I_Validator)it.next();
				b.append(v.toXML(indent + Form.PP_INDENT));
			}
			b.append("\n" + indent + "</"+Form.FBTAG_VALIDATORS+">");
		}
		if(layout != null)
			b.append(layout.toXML(indent)); // don't use method since it may return the settings for the form
		if(getFormat() != null)
			b.append(getFormat().toXML(indent));
		if(b.length() == 0)
 			return null;
 		else
	 		return b.toString();
	}
}
