/*
 * Copyright (c) 2004-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 117 $
 * $LastChangedDate: 2007-02-20 11:24:48 +0100 (Tue, 20 Feb 2007) $
 * 
 */

package com.codedroids.j;

import java.util.*;
import java.io.*;


public class CSVWriter extends Object {
	private BufferedWriter writer;
    private int recordCount;
    private String recordSeparator; // currently ignored
    private String fieldSeparator;
    private String quoteCharacter;
    private String encoding;
    private boolean firstRecordContainsLabels;
    private List labels;
    private Object nullValue;


    public CSVWriter() 
    {
    	this.setRecordSeparator(null);
    	this.setFieldSeparator(null);
        this.setQuoteCharacter(null);
        this.setEncoding(null);
        this.setFirstRecordContainsLabels(true);
        this.setNullValue(null);
    }
    
    public CSVWriter(String recSep, String fieldSep, String quoteChar, String enc, boolean labels, Object nullValue) 
    {
    	this.setRecordSeparator(recSep);
    	this.setFieldSeparator(fieldSep);
        this.setQuoteCharacter(quoteChar);
        this.setEncoding(enc);
        this.setFirstRecordContainsLabels(labels);
        this.setNullValue(nullValue);
    }


    protected void finalize() throws Throwable
    {
        if(writer != null) {
            writer.close();
        }
        super.finalize();
    }
    
    
    public void setRecordSeparator(String recSep)
    {
        if(recSep != null)
            recordSeparator = recSep;
        else
            recordSeparator = java.lang.System.getProperty("line.separator");
    }
    public String recordSeparator()
    {
        return recordSeparator;
    }

    
    public void setFieldSeparator(String fieldSep)
    {
        if(fieldSep != null)
            fieldSeparator = fieldSep;
        else
            fieldSeparator = ",";
    }
    public String fieldSeparator()
    {
        return fieldSeparator;
    }


    public void setQuoteCharacter(String quoteChar)
    {
        if(quoteChar != null)
            quoteCharacter = quoteChar;
        else
            quoteCharacter = "\"";
    }
    public String quoteCharacter()
    {
        return quoteCharacter;
    }
	
    public void setEncoding(String enc)
    {
		encoding = enc;
    }
    public String encoding()
    {
        return encoding;
    }


    public void setFirstRecordContainsLabels(boolean flag)
    {
        firstRecordContainsLabels = flag;
    }
    public boolean firstRecordContainsLabels()
    {
        return firstRecordContainsLabels;
    }


    public void setNullValue(Object nullVal)
    {
        if(nullVal != null)
            nullValue = nullVal;
        else
            nullValue = CSVReader.NullValue; // Same default as in the reader
    }
    public Object nullValue()
    {
        return nullValue;
    }
	
	// convenience that will return the standin null object if passed value is null
	public Object ifNull(Object o) {
		if(o==null)
			return nullValue();
		else
			return o;
	}

    // If labels are set they will determine the which attributes and in what order
    // they appear in the outfile (if input is a dict), also the first row will be the labels.
    public void setLabels(List arr)
    {
        labels = arr;
    }
    public List labels()
    {
        return labels;
    }


    public void open(Writer aWriter) throws IOException
    {
        if(writer != null)
            writer.close();
        if(aWriter instanceof BufferedWriter)
            writer = (BufferedWriter)aWriter;
        else
            writer = new BufferedWriter(aWriter);
        recordCount = 0;
    }
    
	public void open(String aFileName) throws IOException
    {
		if(encoding==null)
			open(new FileWriter(aFileName));
		else
			open(new OutputStreamWriter(new FileOutputStream(aFileName), encoding));
	}
	public void close() throws IOException
    {
        if(writer != null)
            writer.close();
	}

    public String properRecordString(List fields)
    {
        StringBuffer buf = new StringBuffer();
        int cnt = fields.size();
        for(int i=0; i<cnt; i++) {
            if(i>0)
                buf.append(fieldSeparator);
            Object obj = fields.get(i);
            if(obj == nullValue) {
                continue;
            } else {
                String val;
                if(obj instanceof String)
                    val = (String)obj;
                else
                    val = obj.toString();
                boolean quote = false;
                if(val.indexOf(fieldSeparator) >= 0)
                    quote = true;
                else if (val.indexOf(recordSeparator) >= 0)
                    quote = true;
                if(quote)
                    buf.append(quoteCharacter);
                for(int j=0; j<val.length(); j++) {
                    buf.append(val.charAt(j));
                    if(val.charAt(j)==quoteCharacter.charAt(0))
                        buf.append(val.charAt(j));
                }
                if(quote)
                    buf.append(quoteCharacter);
            }
        }
        return new String(buf);
    }
    
    
    public void addRecordString(String rec) throws IOException
    {
        if((recordCount == 0) && firstRecordContainsLabels && (labels != null)) {
            writer.write(this.properRecordString(labels));
            writer.newLine();
            writer.flush();
            recordCount++;
        }
        if(rec != null) {
            writer.write(rec);
            writer.newLine();
            writer.flush();
            recordCount++;
        }
    }
    
    public void addRecordList(List rec) throws IOException
    {
        if(rec != null)
            this.addRecordString(this.properRecordString(rec));
    }
    
    
    public void addRecordMap(Map rec) throws IOException
    {
        if(rec != null) {
            ArrayList arr = new ArrayList();
            if(labels == null)
                labels = new ArrayList(rec.keySet()); // if not specified use all keys as labels

            for(int i=0; i<labels.size(); i++) {
                Object val = rec.get(labels.get(i));
                if(val != null)
                    arr.add(val);
            }
            this.addRecordList(arr);
        }
    }

}

