/*
 * Copyright (c) 2005-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 182 $
 * $LastChangedDate: 2007-07-05 11:13:28 +0200 (Thu, 05 Jul 2007) $
 * 
 */

package com.codedroids.records;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.*;
import java.text.ParseException;
import org.w3c.dom.*;


public class RecordFactory {

	public static final String TAG_BAG_COL  = "bag"; // special 'hidden' column for serialized fields.
	public static final String TAG_CLASS    = "class";
	public static final String TAG_COLUMN   = "column";
	public static final String TAG_ENTITY   = "entity";
	public static final String TAG_FIELD    = "field";
	public static final String TAG_FIELDS   = "fields";
	public static final String TAG_IDENT    = "ident"; // ident field (primary key)
	public static final String TAG_NAME     = "name"; // name of entity/field
	public static final String TAG_TABLE    = "table";
	public static final String TAG_TYPE     = "type";
	public static final String ATTR_SCHEME  = "scheme";
	public static final String ATTR_VERS    = "version";
	
	public static final String REC_ID_SCHEME_SEQ = "sequence"; // using some sequence mechanism provided by subclasses
	public static final String REC_ID_SCHEME_UUID = "uuid"; // using some uuid mechanism provided by subclasses
	public static final String REC_ID_SCHEME_RAND = "random"; // using a random string (using java methods)
	public static final String REC_ID_SCHEME_NONE = "none"; // id must be set before save, simply returns that value
	
	public static final String FILE_EXT = ".rec.xml";
		
	private static HashMap entityFactories = new HashMap();

	public synchronized static RecordFactory getFactory(String name) {
		return (RecordFactory)entityFactories.get(name);
	}

	public synchronized static void setFactory(String name, RecordFactory factory) {
		entityFactories.put(name, factory);
	}
	
	
	protected static void parseEntityFactory(RecordFactory factory, Node root) throws ClassNotFoundException
	{
		NamedNodeMap attrs = root.getAttributes();
		for(int i=0; i<attrs.getLength(); i++) {
			Node attr = attrs.item(i);
			String an = attr.getNodeName();
			if(an.equals(ATTR_VERS)) {
				factory.setVersion(collectText(attr));
			}
		}
		Node child = root.getFirstChild();
		String ident = null;
		while( child != null) {
			short nt = child.getNodeType();
			switch(nt) {
				case Node.ELEMENT_NODE:
					String nn = child.getNodeName();
					if (nn.equals(TAG_NAME)) {
						factory.setName(collectText(child));
					}
					else if (nn.equals(TAG_TABLE)) {
						factory.setTableName(collectText(child));
					}
					else if (nn.equals(TAG_CLASS)) {
						factory.setInstanceClass(collectText(child));
					}
					else if (nn.equals(TAG_IDENT)) {
						ident = collectText(child);
						attrs = child.getAttributes();
						for(int i=0; i<attrs.getLength(); i++) {
							Node attr = attrs.item(i);
							String an = attr.getNodeName();
							if(an.equals(ATTR_SCHEME)) {
								factory.setIdentScheme(collectText(attr));
							}
						}
					}
					else if (nn.equals(TAG_BAG_COL)) {
						factory.setBagColumn(collectText(child));
					}
					else if (nn.equals(TAG_FIELDS)) {
						parseEntityFields(factory, child);
					}
					break;
				default:		
					break;
			}
			child = child.getNextSibling();
		}
		factory.setIdentField(factory.getField(ident));
	}
	
	protected static void parseEntityFields(RecordFactory factory, Node nodes)
	{
		NodeList nl = nodes.getChildNodes();
		int len = nl.getLength();
		for(int i = 0; i < len; i++) {
			Node field = nl.item(i);
			short ft = field.getNodeType();
			switch(ft) {
				case Node.ELEMENT_NODE:
					String fn = field.getNodeName();
					if (fn.equals(TAG_FIELD)) {
						NodeList children = field.getChildNodes();
						RecordField ef = new RecordField();
						for(int j=0; j<children.getLength(); j++) {
							Node fieldAttr = children.item(j);
							short fat = field.getNodeType();
							switch(fat) {
								case Node.ELEMENT_NODE:
									String nn = fieldAttr.getNodeName();
									if (nn.equals(TAG_NAME)) {
										ef.setName(collectText(fieldAttr));
									}
									else if (nn.equals(TAG_COLUMN)) {
										ef.setColumn(collectText(fieldAttr));
									}
									else if (nn.equals(TAG_TYPE)) {
										ef.setType(collectText(fieldAttr));
									}
									break;
								default:		
									break;
							}
						}
						factory.addField(ef);
					}
					break;
				default:		
					break;
			}
		}
	}
	
	/**
	 * Collect text by concatenating all TEXT and CDATA-nodes into string
	 * @param node
	 * @return String containing the text or null
	 */
	public static String collectText(Node node) {
		Node child = node.getFirstChild();
		StringBuffer retval = new StringBuffer();
		while(child != null) {
			switch(child.getNodeType()) {
				case Node.TEXT_NODE:
				case Node.CDATA_SECTION_NODE:
					retval.append(child.getNodeValue());
			}
			child = child.getNextSibling();
		}
		if(retval.length()==0)
			return null;
		else
			return retval.toString();
	}
	
	//
	// Instance 
	//
	
	private Class instanceClass;
	private String name;
	private String tableName;
	private RecordField identField;
	private String bagColumn;
	private Map fields;
	private String scheme; 
	private String version; 
	private MySQLGenerator sql; // for now...
	
	public RecordFactory() {
		super();
		fields = new TreeMap();
		// we can do this as long as the generator does not try to interrogate us before the xml is loaded
		sql = new MySQLGenerator(this); 
	}
	
	public Class getInstanceClass() {
		return instanceClass;
	}
	public void setInstanceClass(String className) throws ClassNotFoundException {
		setInstanceClass(Class.forName(className));
	}
	public void setInstanceClass(Class instanceClass) {
		this.instanceClass = instanceClass;
	}

	public String getName() {
		return name;
	}
	public void setName(String entityName) {
		this.name = entityName;
	}
	
	public String getVersion() {
		if(version==null)
			return "1";
		else
			return version;
 	}

	public void setVersion(String str) {
		this.version = str;
	}
	
	public String getIdentScheme() {
		if(scheme==null)
			return REC_ID_SCHEME_SEQ;
		else
			return scheme;
 	}

	public void setIdentScheme(String str) {
		this.scheme = str;
	}
	
	public String getTableName() {
		return tableName;
	}
	public void setTableName(String tableName) {
		this.tableName = tableName;
	}
	
	public RecordField getIdentField() {
		return identField;
	}
	public void setIdentField(RecordField identField) {
		this.identField = identField;
	}
	
	public String getBagColumn() {
		return bagColumn;
	}
	public void setBagColumn(String bagColumn) {
		this.bagColumn = bagColumn;
	}
	
	public Map getFields() {
		return fields;
	}

	public void setFields(Map fields) {
		this.fields = fields;
	}
	
	public void addField(RecordField f) {
		if(this.fields != null)
			this.fields.put(f.getName(), f);
	}
	
	public void removeField(RecordField f) {
		if(this.fields != null)
			this.fields.remove(f.getName());
	}
	
	public RecordField getField(String fieldName) {
		return (RecordField)fields.get(fieldName);
	}

	//
	// I_Record instance handling
	//
	
	public I_Record getNewInstance() throws RecordException {
		try {
			I_Record e = (I_Record)instanceClass.newInstance();
			e.init(this);
			return e;
		} catch (Exception e) {
			throw new RecordException("Failed to get new instance", e);
		}
	}
	
	public I_Record getNewInstance(String id) throws RecordException {
		try {
			I_Record e = (I_Record)instanceClass.newInstance();
			e.init(this, getIdentField().fromString(id));
			return e;
		} catch (Exception e) {
			throw new RecordException("Failed to get new instance", e);
		}
	}
	
	public I_Record getNewInstance(Object id) throws RecordException {
		try {
			I_Record e = (I_Record)instanceClass.newInstance();
			e.init(this, id);
			return e;
		} catch (Exception e) {
			throw new RecordException("Failed to get new instance", e);
		}
	}
	
	public I_Record getInstanceCopy(I_Record original) throws RecordException {
		I_Record copy = getNewInstance();
		return getInstanceCopy(original, copy);
	}
	
	public I_Record getInstanceCopy(I_Record original, I_Record copy) {
		Iterator i = fields.values().iterator();
		while(i.hasNext()) {
			RecordField f = (RecordField)i.next();
			if(!f.equals(getIdentField()))
				copy.takeValueForField(f, original.valueForField(f));
		}
		return copy;
	}
	
	public void populateFromResultSet(I_Record record, ResultSet rs) throws RecordException {
		try {
			Iterator keys = getFields().keySet().iterator();
			int index = 1;
			while (keys.hasNext()) {
				RecordField field = (RecordField)getFields().get((String)keys.next());
				record.takeValueForField(field, field.getFromResultSet(rs, index++));
			}
			record.fetched(); // mark as not new.
			//A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[populateFromResultSet] =" + record.toString());
		} catch (Exception e) {
			throw new RecordException("Failed to populare record from result set", e);
		}
	}
	
	public void pushToPreparedStatement(I_Record record, PreparedStatement ps, boolean update) throws RecordException {
		try {
			Iterator keys = getFields().keySet().iterator();
			int index = 1;
			while (keys.hasNext()) {
				RecordField field = (RecordField)getFields().get((String)keys.next());
				if(update && field==getIdentField())
					continue; // don't update id
				field.setOnPreparedStatement(ps, index++, record.valueForField(field));
			}
			if(update)
				getIdentField().setOnPreparedStatement(ps, index, record.valueForField(getIdentField()));
			//A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[pushToPreparedStatement] =" + ps);
		} catch (Exception e) {
			throw new RecordException("Failed to push record to prepared statement", e);
		}
	}

	//
	// Subclasses should override these (should be made abstract at some point...)
	//
	
	public String getPoolName() {
		return null;
	}

	/**
	 * @deprecated use the one below that takes the record as param.
	 */
	public Integer getNextId(String tableName) {
		return null;
	}

	public Object getNextId(I_Record record) {
		return null;
	}
	
	public void clearCache() {
		return;
	}

	public void debug(String msg) {
		System.out.println(msg);
	}
	
	//
	// These does the hard work.
	//
	
	/**
	 * Fetch all records from table
	 * 
	 * @return a list (possibly empty, but not null)
	 */
	public List fetchRecords() throws RecordException {
		return fetchRecords(null, null);
	}
	/**
	 * Fetch all records from table matching the given where clause
	 * 
	 * @return a list (possibly empty, but not null)
	 */
	public List fetchRecords(A_Clause whereClause) throws RecordException {
		return fetchRecords(whereClause, null);
	}
	/**
	 * Fetch all records from table matching the given where clause, ordered by the given specification.
	 * 
	 * @return a list (possibly empty, but not null)
	 */
	public List fetchRecords(A_Clause whereClause, Object orderBy) throws RecordException  {
		List list = new ArrayList();
        Connection con = null;
		PreparedStatement statement = null;
		ResultSet res = null;
		
		try {
            con = DriverManager.getConnection(getPoolName());
			statement = con.prepareStatement(sql.selectAllSQL(whereClause, orderBy));
			if(whereClause != null && whereClause.cardinality() > 0) {
				List params = whereClause.getValues();
				List paramFields = whereClause.getFields();
				for(int ndx = 0; ndx < params.size(); ndx++) {
					RecordField f = (RecordField)paramFields.get(ndx);
				    f.setOnPreparedStatement(statement, ndx+1, params.get(ndx));
				}
			}
//			debug("[FETCH RECORDS] = " + statement);
			res = statement.executeQuery();
			while(res.next()){
				I_Record record = getNewInstance();
				populateFromResultSet(record, res);
				list.add(record);
			}
		} catch (Exception e) {
			throw new RecordException("Failed to fetch records where: \"" + whereClause + "\"", e);
		} finally {
			try{
				res.close();
			} catch (Exception e) {
				// ignore the exception
			}
            try{
				statement.close();
			} catch (Exception e) {
				// ignore the exception
			}
            try{
				con.close();
			} catch (Exception e) {
				// ignore the exception
			}
		}
		return list;
	}
	
	public I_Record fetchRecord(String id) throws RecordException, ParseException {
		return fetchRecord(getIdentField().fromString(id));
	}
	
	public I_Record fetchRecord(Object id) throws RecordException {
		I_Record record = null;
		record = fetchRecord(id, getNewInstance());
		return record;
	}
	
	public I_Record fetchRecord(Object id, I_Record record) throws RecordException {
		return fetchRecord(new GenericClause(getIdentField(),"=",id), record);
	}
	
	public I_Record fetchRecord(A_Clause whereClause) throws RecordException {
		I_Record record = null;
		record = fetchRecord(whereClause, getNewInstance());
		return record;
	}
	
	public I_Record fetchRecord(A_Clause whereClause, I_Record record) throws RecordException {
		Connection con = null;
		PreparedStatement statement = null;
		ResultSet res = null;
		try {
			con = DriverManager.getConnection(getPoolName());
			statement = con.prepareStatement(sql.selectAllSQL(whereClause, null));
			if(whereClause != null && whereClause.cardinality() > 0) {
				List params = whereClause.getValues();
				List paramFields = whereClause.getFields();
				for(int ndx = 0; ndx < params.size(); ndx++) {
					RecordField f = (RecordField)paramFields.get(ndx);
				    f.setOnPreparedStatement(statement, ndx+1, params.get(ndx));
				}
			}
//			debug("[FETCH RECORD] = " + statement);
			res = statement.executeQuery();
			if (res.next()) {
				populateFromResultSet(record, res); // ignore if more than one record selected
			} else {
				record = null;
			}
		} catch (Exception e) {
			throw new RecordException("Failed to fetch record where: \"" + whereClause + "\"", e);
		} finally {
			try{
				res.close();
			} catch (Exception e) {
				// ignore the exception
			}
			try{
				statement.close();
			} catch (Exception e) {
				// ignore the exception
			}
			try{
				con.close();
			} catch (Exception e) {
				// ignore the exception
			}
		}
		return record;
	}
	
	public void deleteRecord(I_Record rec) throws RecordException, ParseException {
		deleteRecord(rec.valueForField(getIdentField()));
	}

	public void deleteRecord(String id) throws RecordException, ParseException {
		deleteRecord(getIdentField().fromString(id));
	}
	
	public void deleteRecord(Object id) throws RecordException {
		Connection con = null;
		PreparedStatement statement = null;
		//ResultSet res = null;
		clearCache();
		try {
			con = DriverManager.getConnection(getPoolName());
			statement = con.prepareStatement(sql.deleteEntrySQL());
			getIdentField().setOnPreparedStatement(statement, 1, id);
//			debug("[DELETE RECORD] = " + statement);
			statement.execute();
		} catch (Exception e) {
			throw new RecordException("Failed to delete record with ident: \"" + id + "\"", e);
		} finally {
			 try{
				statement.close();
			}catch (Exception e) {
				// ignore the exception
			}
			try{
				con.close();
			}catch (Exception e) {
				// ignore the exception
			}
		}
	}

	public void saveRecord(I_Record record) throws RecordException 
	{
		Connection con = null;
		PreparedStatement statement = null;
		//ResultSet res = null;
		clearCache(); // clear the cache to make changes visible
		try {
			if(record.isNew()) {
				record.takeValueForField(this.getIdentField(), getNextId(record));
				con = DriverManager.getConnection(getPoolName());
				statement = con.prepareStatement(sql.insertEntrySQL());
				pushToPreparedStatement(record, statement, false);
//				debug("[WRITE RECORD] + " + statement.toString());
				statement.execute();
				record.saved();
			} else {
				con = DriverManager.getConnection(getPoolName());
				statement = con.prepareStatement(sql.updateEntrySQL());
				pushToPreparedStatement(record, statement, true);
//				debug("[UPDATE RECORD] + " + statement.toString());
				statement.execute();
			}
		} catch (Exception e) {
			throw new RecordException("Failed to save" + (record.isNew() ? " new " : " ") + "record: \"" + record + "\"", e);
		} finally {
			try{
				statement.close();
			}catch (Exception e) {
				// ignore the exception
			}
			try{
				con.close();
			}catch (Exception e) {
				// ignore the exception
			}
		}
	}
	
	//
	// Stuff
	//
	
	public String toString() {
		return toXML();
	}

	private static String PP = "    ";
	public String toXML() {
		StringBuffer b = new StringBuffer();
		b.append("\n<"+TAG_ENTITY+" "+ATTR_VERS+"=\""+getVersion()+"\">");
		b.append("\n"+PP+"<"+TAG_NAME+">" + getName() + "</"+TAG_NAME+">");
		b.append("\n"+PP+"<"+TAG_CLASS+">" + getInstanceClass().getName() + "</"+TAG_CLASS+">");
		b.append("\n"+PP+"<"+TAG_TABLE+">" + getTableName() + "</"+TAG_TABLE+">");
		b.append("\n"+PP+"<"+TAG_IDENT+">" + getIdentField().getName() + "</"+TAG_IDENT+">");
		b.append("\n"+PP+"<"+TAG_BAG_COL+">" + getBagColumn() + "</"+TAG_BAG_COL+">");
		if(getFields() != null) {
			b.append("\n"+PP+"<"+TAG_FIELDS+">");
			Iterator it = getFields().keySet().iterator();
			while(it.hasNext()){
				RecordField ef = getField((String)it.next());
				b.append("\n"+PP+PP+"<"+TAG_FIELD+">");
				b.append("\n"+PP+PP+PP+"<"+TAG_NAME+">" + ef.getName() + "</"+TAG_NAME+">");
				b.append("\n"+PP+PP+PP+"<"+TAG_TYPE+">" + ef.getType() + "</"+TAG_TYPE+">");
				b.append("\n"+PP+PP+PP+"<"+TAG_COLUMN+">" + ef.getColumn() + "</"+TAG_COLUMN+">");
				b.append("\n"+PP+PP+"</"+TAG_FIELD+">");
			}
			b.append("\n"+PP+"</"+TAG_FIELDS+">");
		}
		b.append("\n</"+TAG_ENTITY+">");
		return b.toString();
	}
}
